
local AbsorbsMonitor = _G.AbsorbsMonitor;

local Scaling;
local playerScaling;
local privateScaling;

local playerGUID;
local playerClass;


-- #################################################################################################


-- No function in this file will be called before this
function AbsorbsMonitor:EnableEffects()
	Scaling = AbsorbsMonitor.Scaling;
	playerGUID = UnitGUID("player");
	_, playerClass = UnitClass("player");
	
	playerScaling = Scaling[playerGUID];
	privateScaling = Scaling[-1];
end


-- #################################################################################################


DEATHKNIGHT_ANTIMAGICSHELL_SPELLID = 48707

-- Public Scaling:
--   {[DEATHKNIGHT_ANTIMAGICSHELL_SPELLID] = Magic Suppression Factor}

local deathknight_MS_Ranks = {[0] = 0, [1] = 0.08, [2] = 0.16, [3] = 0.25};

function deathknight_AntiMagicShell_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	local maxHealth = UnitHealthMax(destName);
	
	if(maxHealth) then		
		return math.floor(maxHealth * 0.5), (Scaling[sourceGUID] and Scaling[sourceGUID][DEATHKNIGHT_ANTIMAGICSHELL_SPELLID]) and 1.0 or 0.5;
	else
		return 0, 0.0;
	end
end

local function deathknight_AntiMagicShell_Hit(guid, curEffect, absorbedRemaining, overkill, spellSchool) 
	-- TODO: what happens to mixed school attacks?
	if(spellSchool == SCHOOL_MASK_PHYSICAL) then 
		return 0, true;
	end
	
	-- TODO: make it possible to store custom data with an active absorb effect
	local maxAbsorb = (absorbedRemaining + overkill) * (0.75 + (Scaling[guid] and Scaling[guid][DEATHKNIGHT_ANTIMAGICSHELL_SPELLID] or 0));
	
	if(curEffect[3] < maxAbsorb) then
		return curEffect[3], false;
	else
		return maxAbsorb, true;
	end
end

local function deathknight_OnTalentUpdate(self)
	local _, _, _, _, t = GetTalentInfo(3, 18);
	
	playerScaling[DEATHKNIGHT_ANTIMAGICSHELL_SPELLID] = deathknight_MS_Ranks[t];
end

function AbsorbsMonitor:OnEnable_DEATHKNIGHT()
	self.OnTalentUpdate = deathknight_OnTalentUpdate;
	
	deathknight_OnTalentUpdate();
end


-- #################################################################################################


local function druid_SavageDefense_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	if(sourceGUID == playerGUID) then
		local baseAP, plusAP, minusAP = UnitAttackPower("player");
	
		return math.floor((baseAP + plusAP - minusAP) * 0.25), 1.0;
	else
		return 0, 0.0
	end
end

local function druid_SavageDefense_Hit(guid, curEffect, absorbedRemaining, overkill, spellSchool) 
	-- TODO: what happens to mixed school attacks?
	if(spellSchool == SCHOOL_MASK_PHYSICAL) then 
		return min(curEffect[3], absorbedRemaining), false;
	else
		return 0, true;
	end
	
	
end


-- #################################################################################################


local function paladin_SacredShield_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	return 0, 0.0;
end


-- #################################################################################################


PRIEST_DIVINEAEGIS_SPELLID = 47753;

-- [rank] = {spellId, level, value}
-- Value will be extracted from the spell book by priest_ScanSpellbook()
local priest_PWS_Ranks = {
	[1] = {17, 6, 0},
	[2] = {592, 12, 0},
	[3] = {600, 18, 0},
	[4] = {3747, 24, 0},
	[5] = {6065, 30, 0},
	[6] = {6066, 36, 0},
	[7] = {10898, 42, 0},
	[8] = {10899, 48, 0},
	[9] = {10900, 54, 0},
	[10] = {10901, 60, 0},
	[11] = {25217, 65, 0},
	[12] = {25218, 70, 0},
	[13] = {48065, 75, 0},
	[14] = {48066, 80, 0}
};

-- Public Scaling:
--   Power Word: Shield: [spellId] = {base, spFactor}
--   Divine Aegis: [47753] = healFactor

-- Private Scaling
--   Talents: "TwinDisc", "ImpPWS", "FocusedPower", "DivineAegis", "BorrowedTime", "SpiritualHealing"
--	 Gear: "4pcRaid9", "4pcRaid10"

local function priest_PowerWordShield_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	local value;
	
	if(sourceGUID == playerGUID) then
		return math.floor((playerScaling[spellId][1] + GetSpellBonusHealing() * playerScaling[spellId][2]) * AbsorbsMonitor.db.profile.priest_globalFactor), 1.0;
	else
		return 0, 0.0;
	end
end

local function priest_DivineAegis_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)		
	local existing = 0;
	
	if(destEffects and destEffects[spellId]) then
		existing = destEffects[spellId][3];
	end

	local charged = AbsorbsMonitor:ConsumeCharges(destGUID, spellId);
	
	if(charged == 0) then
		return existing, 0.0;
	end
	
	local destLevel = UnitLevel(destName);
	local quality = 1.0;		
	
	if(not destLevel) then
		destLevel = 80;
		quality = 0.4;
	end
		
	return min(destLevel * 125, existing + charged), quality;
end

local function priest_DivineAegis_OnHealCrit(sourceGUID, sourceName, destGUID, destName, spellId, amount)	
	-- No need for any check anymore, since the callback is only in place
	-- if the source has Divine Aegis
	
	-- The latency is horrible sometimes, even 5.0 might be too low
	-- Should be modified by 
	AbsorbsMonitor:AddCharge(destGUID, PRIEST_DIVINEAEGIS_SPELLID, math.floor(amount * playerScaling[PRIEST_DIVINEAEGIS_SPELLID]), 5.0);
end

local function priest_CalculateScaling()
	local playerLevel_minus5 = UnitLevel("player") - 5;
	local base_spFactor = 0.807;
	
	base_spFactor = base_spFactor + (privateScaling["BorrowedTime"] * 0.08);
	base_spFactor = base_spFactor * (1.0 + (privateScaling["TwinDisc"] * 0.01) + (privateScaling["FocusedPower"] * 0.02) + (privateScaling["SpiritualHealing"] * 0.02)) * (1.0 + ((privateScaling["ImpPWS"] + privateScaling["4pcRaid10"]) * 0.05));	
	
	playerScaling[PRIEST_DIVINEAEGIS_SPELLID] = (privateScaling["DivineAegis"] * 0.1) * (1 + (privateScaling["4pcRaid9"] * 0.03));
	
	if(playerScaling[PRIEST_DIVINEAEGIS_SPELLID] > 0) then
		AbsorbsMonitor:AddCustomTrigger(playerGUID, "OnHealCrit", priest_DivineAegis_OnHealCrit);
	else
		AbsorbsMonitor:RemoveCustomTrigger(playerGUID, "OnHealCrit", priest_DivineAegis_OnHealCrit);
	end
	
	local spFactor;
	
	for k, v in pairs(priest_PWS_Ranks) do
		if(v[3]) then
			-- Based on the assumption that the decrease in sp coefficient is linear,
			-- only tested for level 80 though so far
			-- Cataclysm will help us get rid of this crap again
			if(v[2] < playerLevel_minus5) then
				if(playerLevel_minus5 == 75) then
					spFactor = max(base_spFactor * ((v[2] * 0.045228921) - 2.381389768), 0);
				else
					spFactor = 0;
				end
			else
				spFactor = base_spFactor;
			end
			
			playerScaling[v[1]] = {v[3] * (1.0 + (privateScaling["TwinDisc"] * 0.01) + (privateScaling["FocusedPower"] * 0.02) + (privateScaling["SpiritualHealing"] * 0.02)), spFactor};
		end
	end
end

local function priest_ScanSpellbook()
	local PWS_Tooltip = {};
	PWS_Tooltip.t = CreateFrame("GameTooltip");
	PWS_Tooltip.t:SetOwner(UIParent, "ANCHOR_NONE");
	
	for i = 1, 4 do
		PWS_Tooltip["TooltipTextLeft" .. i] = PWS_Tooltip.t:CreateFontString();
		PWS_Tooltip["TooltipTextRight" .. i] = PWS_Tooltip.t:CreateFontString();
		PWS_Tooltip.t:AddFontStrings(PWS_Tooltip["TooltipTextLeft" .. i], PWS_Tooltip["TooltipTextRight" .. i]);
	end
		
	local spellName, spellRank, spellId;
	local i = 1;
	
	while true do
		spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL);
		
		if(not spellName) then break; end
		
		if(spellName == "Power Word: Shield") then
			spellRank = tonumber(spellRank:match("(%d+)"));			
			spellId = priest_PWS_Ranks[spellRank][1];
			
			if(not spellId) then
				self:Print("could not found spell id for rank "..spellRank);
				break;
			end						
			
			PWS_Tooltip.t:SetSpell(i, BOOKTYPE_SPELL);
			
			local _, _, value = string.find(PWS_Tooltip.TooltipTextLeft4:GetText(), "absorbing (%d+) damage");						
			
			priest_PWS_Ranks[spellRank][3] = value;			
		end
		
		i = i + 1;
	end	
	
	PWS_Tooltip.t:SetParent(nil);
	PWS_Tooltip = nil;
end

local function priest_ScanTalents()
	-- Twin Disciplines
	local _, _, _, _, t = GetTalentInfo(1, 2);
	privateScaling["TwinDisc"] = t;
	
	-- Improved Power Word: Shield
	local _, _, _, _, t = GetTalentInfo(1, 7);
	privateScaling["ImpPWS"] = t;
	
	-- Focused Power
	local _, _, _, _, t = GetTalentInfo(1, 16);
	privateScaling["FocusedPower"] = t;
	
	-- Divine Aegis
	local _, _, _, _, t = GetTalentInfo(1, 24);		
	privateScaling["DivineAegis"] = t;
	
	-- Borrowed Time
	local _, _, _, _, t = GetTalentInfo(1, 27);
	privateScaling["BorrowedTime"] = t;
	
	-- Spiritual Healing
	local _, _, _, _, t = GetTalentInfo(2, 16);
	privateScaling["SpiritualHealing"] = t;
end

local function priest_ScanEquipment()
	local n = 0;
	
	-- Crimson Acolyte Raiment's 4-piece Bonus
	if(IsEquippedItem("Crimson Acolyte Hood") or IsEquippedItem("Sanctified Crimson Acolyte Hood")) 				then n = 1; end
	if(IsEquippedItem("Crimson Acolyte Shoulderpads") or IsEquippedItem("Sanctified Crimson Acolyte Shoulderpads")) then n = n + 1; end
	if(IsEquippedItem("Crimson Acolyte Robe") or IsEquippedItem("Sanctified Crimson Acolyte Robe")) 				then n = n + 1; end
	if(IsEquippedItem("Crimson Acolyte Gloves") or IsEquippedItem("Sanctified Crimson Acolyte Gloves")) 			then n = n + 1; end
	if(IsEquippedItem("Crimson Acolyte Leggings") or IsEquippedItem("Sanctified Crimson Acolyte Leggings")) 		then n = n + 1; end
	
	if(n >= 4) then
		privateScaling["4pcRaid10"] = 1;
		
		 -- no way to have 4pcRaid9 now
		privateScaling["4pcRaid9"] = 0;		
		return;
	else
		privateScaling["4pcRaid10"] = 0;
	end
	
	n = 0;
	
	if(UnitFactionGroup("player") == "Alliance") then
		if(IsEquippedItem("Velen's Cowl of Conquest") or IsEquippedItem("Velen's Cowl of Triumph")) 				then n = 1; end
		if(IsEquippedItem("Velen's Shoulderpads of Conquest") or IsEquippedItem("Velen's Shoulderpads of Triumph")) then n = n + 1; end
		if(IsEquippedItem("Velen's Robe of Conquest") or IsEquippedItem("Velen's Robe of Triumph")) 				then n = n + 1; end
		if(IsEquippedItem("Velen's Gloves of Conquest") or IsEquippedItem("Velen's Gloves of Triumph")) 			then n = n + 1; end
		if(IsEquippedItem("Velen's Leggings of Conquest") or IsEquippedItem("Velen's Leggings of Triumph")) 		then n = n + 1; end
	else
		if(IsEquippedItem("Zabra's Cowl of Conquest") or IsEquippedItem("Zabra's Cowl of Triumph")) 				then n = 1; end
		if(IsEquippedItem("Zabra's Shoulderpads of Conquest") or IsEquippedItem("Zabra's Shoulderpads of Triumph")) then n = n + 1; end
		if(IsEquippedItem("Zabra's Robe of Conquest") or IsEquippedItem("Zabra's Robe of Triumph")) 				then n = n + 1; end
		if(IsEquippedItem("Zabra's Gloves of Conquest") or IsEquippedItem("Zabra's Gloves of Triumph")) 			then n = n + 1; end
		if(IsEquippedItem("Zabra's Leggings of Conquest") or IsEquippedItem("Zabra's Leggings of Triumph")) 		then n = n + 1; end	
	end
	
	if(n >= 4) then
		privateScaling["4pcRaid9"] = 1;
	else
		privateScaling["4pcRaid9"] = 0;
	end
end

--[[
-- Come on, this whole stuff is bullshit anyway
-- and will hopefully be removed in Cataclysm
local function priest_ScanZoneBuff()
	if(UnitInBattleground("player")) then
		GlobalModifier = 0.9;
	elseif(IsActiveBattlefieldArena()) then
		GlobalModifier = 0.9;
	elseif(GetRealZoneText() == "Wintergrasp") then
		GlobalModifier = 0.9;
	elseif(GetRealZoneText() == "Icecrown Citadel") then
		GlobalModifier = 1.1;
	else
		GlobalModifier = 1.1;
	end
end
]]

local function priest_OnLevelUp(self)
	priest_ScanSpellbook();
	priest_CalculateScaling();
end

local function priest_OnTalentUpdate(self)
	priest_ScanTalents();
	priest_CalculateScaling();
end

local function priest_OnEquipmentChangedDelayed(self)
	priest_ScanEquipment();	
	priest_CalculateScaling();
	
	self:FinishUniqueTimer("OnEquipChanged");
end

local function priest_OnEquipmentChanged(self)
	self:ScheduleUniqueTimer("OnEquipChanged", priest_OnEquipmentChangedDelayed, 0.7, self);
end

function AbsorbsMonitor:OnEnable_PRIEST()
	self.OnLevelUp = priest_OnLevelUp;
	self.OnTalentUpdate = priest_OnTalentUpdate;
	self.OnEquipmentChanged = priest_OnEquipmentChanged;
	
	priest_ScanSpellbook();
	priest_ScanTalents();
	priest_ScanEquipment();
	
	priest_CalculateScaling();
end


-- #################################################################################################


local function items_HitEssenceOfGossamer(guid, name, curEffect)
	if(curEffect[3] < 140) then
		return curEffect[3], false;
	else
		return 140, true;
	end
end

local function items_Valanyr_OnAuraApplied(sourceGUID, sourceName, destGUID, destName, spellId)
	AbsorbsMonitor:AddCustomTrigger(sourceGUID, "OnHeal", items_Valanyr_OnHeal);
end

local function items_Valanyr_OnAuraRemoved(sourceGUID, sourceName, destGUID, destName, spellId)
	AbsorbsMonitor:RemoveCustomTrigger(sourceGUID, "OnHeal", items_Valanyr_OnHeal);
end

local function items_Valanyr_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)		
	local existing = 0;
	
	if(destEffects and destEffects[spellId]) then
		existing = destEffects[spellId][3];
	end

	local charged = AbsorbsMonitor:ConsumeCharges(destGUID, spellId);
	
	if(charged == 0) then
		return existing, 0.0;
	end
		
	return min(20000, existing + charged), 1.0;
end

local function items_Valanyr_OnHeal(sourceGUID, sourceName, destGUID, destName, spellId, amount)
	AbsorbsMonitor:AddCharge(destGUID, 64413, math.floor(amount * 0.15), 5.0);
end


-- #################################################################################################


--- Generic Create function (only for documentary purposes)
-- @param	sourceGUID			guid of the originating unit
-- @param	sourceName			name of the originating unit
-- @param	destGUID			guid of the affected unit
-- @param	destName			name of the affected unit
-- @param	spellId				the spellId of this absorb effect
-- @param	destEffects			activeEffectsBySpell[destGUID]
-- @return	total value of this absorb effect
-- @return	quality (that is accuracy) of this value
local function generic_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
end

--- Generic Hit function suitable for most absorb effects
-- Note that this function is only responsible for determining the amount this
-- particular absorb effect will take, NOT to handle its consequences like updating
-- the data structures
-- @param	guid				guid of affected unit
-- @param	curEffect			the value of activeEffectsBySpell[guid][spellId]
-- @param	absorbedRemaining	absorb value left to be accounted for on this unit
-- @param	overkill			amount of damage done on top of the absorb
-- @param	spellSchool			spell school for this hit
-- @return	absorb value this effect can account for
-- @return	whether this absorb was broken by this hit
local function generic_Hit(guid, curEffect, absorbedRemaining, overkill, spellSchol)
	if(absorbedRemaining > curEffect[3]) then
		-- dirty but efficient
		absorbedRemaining = curEffect[3];
		overkill = 1;
	end

	return absorbedRemaining, (overkill == 0);
end


-- #################################################################################################


local PWS_Entry = {1.0, priest_PowerWordShield_Create, generic_Hit};	
local unsupportedEntry = {1, function() return 0, 0.0; end, generic_Hit};

-- INCOMPLETE
-- This table gets a lot smaller in Cataclysm
AbsorbsMonitor.Effects = {
	-- DEATH KNIGHT	
	-- Anti-Magic Shell
	[48707] = {3.0, deathknight_AntiMagicShell_Create, deathknight_AntiMagicShell_Hit},
		
	-- DRUID
	-- Savage Defense
	[62606] = {1.1, druid_SavageDefense_Create, druid_SavageDefense_Hit},
	-- Sacred Shield
	[58597] = {1.0, paladin_SacredShield_Create, generic_Hit},
	
	-- MAGE
	-- Fire Ward
	[543] = unsupportedEntry,
	[8457] = unsupportedEntry,
	[8458] = unsupportedEntry,
	[10223] = unsupportedEntry,
	[10225] = unsupportedEntry,
	[27218] = unsupportedEntry,
	[43010] = unsupportedEntry,		
	-- Frost Ward
	[6143] = unsupportedEntry,
	[8461] = unsupportedEntry,
	[8462] = unsupportedEntry,
	[10177] = unsupportedEntry,
	[28609] = unsupportedEntry,
	[32796] = unsupportedEntry,
	[43012] = unsupportedEntry,		
	-- Ice Barrier
	[11426] = unsupportedEntry,
	[13031] = unsupportedEntry,
	[13032] = unsupportedEntry,
	[13033] = unsupportedEntry,
	[27134] = unsupportedEntry,
	[33405] = unsupportedEntry,
	[43038] = unsupportedEntry,
	[43039] = unsupportedEntry,			
	-- Mana Shield
	[1463] = unsupportedEntry,
	[8494] = unsupportedEntry,
	[8495] = unsupportedEntry,
	[10191] = unsupportedEntry,
	[10192] = unsupportedEntry,
	[10193] = unsupportedEntry,
	[27131] = unsupportedEntry,
	[43019] = unsupportedEntry,
	[43020] = unsupportedEntry,		
	
	-- PALADIN
	-- Sacred Shield
	[58597] = {1.0, paladin_SacredShield_Create, generic_Hit},
	
	-- PRIEST	
	-- Power Word: Shield
	[17] 	= PWS_Entry,
	[592]	= PWS_Entry,
	[600] 	= PWS_Entry,
	[3747] 	= PWS_Entry,
	[6065] 	= PWS_Entry,
	[10898] = PWS_Entry,
	[10899] = PWS_Entry,
	[10900] = PWS_Entry,
	[10901] = PWS_Entry,
	[25217] = PWS_Entry,
	[25218] = PWS_Entry,
	[48065] = PWS_Entry,
	[48066] = PWS_Entry,
	-- Divine Aegis
	[47753] = {1.0, priest_DivineAegis_Create, generic_Hit},
	
	-- ROGUE
	-- Cheat Death
	--   No clue how it works technically (does it still uses an absorb?)
	--   While there is no absorb value to watch for, it would screw our
	--   numbers on other effects.
	
	-- SHAMAN
	-- Glyph of Stoneclaw Totem
	--   This absorb effect cannot be supported without heavy exceptional
	--   handling, because it is neither applied by SPELL_AURA_APPLIED or
	--   nor removed by SPELL_AURA_REMOVED - but only an UNIT_AURA event is
	--   fired on gain/loss. The gain can actually be detected by the cast 
	--   of Stoneclaw Totem (and knowledge about the glyphs), but not if it
	--   breaks prematurely, making looping through all buffs needed everytime
	--	 it could be active and UNIT_AURA is fired. At least the amount would
	--	 be constant for every rank.
	-- The normal totem absorb effect could be tracked with more ease, but still
	-- there is no proper removal event, so it would rely on zombified absorb
	-- effects until the totem drops. Maybe later. (What about totem GUIDs?)
	-- [55277] = {1.0, shaman_GlyphStoneclawTotem_Create, generic_Hit},
	
	-- WARLOCK
	-- Sacrifice
	[7812] = unsupportedEntry,
	[19438] = unsupportedEntry,
	[19440] = unsupportedEntry,
	[19441] = unsupportedEntry,
	[19442] = unsupportedEntry,
	[19443] = unsupportedEntry,
	[27273] = unsupportedEntry,
	[47985] = unsupportedEntry,
	[47986] = unsupportedEntry,
	-- Shadow Ward
	[6229] = unsupportedEntry,
	[11739] = unsupportedEntry,
	[11740] = unsupportedEntry,
	[28610] = unsupportedEntry,
	[47890] = unsupportedEntry,
	[47891] = unsupportedEntry,		

	-- ITEMS	
	-- Val'anyr
	[64413] = {1.0, items_Valanyr_Create, generic_Hit},
	-- Essence of Gossamer
	[60218] = {5.0, function() return 4000, 1.0; end, items_EssenceOfGossamer_Hit},
};

AbsorbsMonitor.Triggers = {	
	-- Val'anyr
	["64411_OnAuraApplied"] = items_Valanyr_OnAuraApplied,
	["64411_OnAuraRemoved"] = items_Valanyr_OnAuraRemoved,
};
