
local AbsorbsMonitor = _G.AbsorbsMonitor;

local AceConfig = LibStub("AceConfig-3.0");
local AceConfigDialog = LibStub("AceConfigDialog-3.0");
local AceDB = LibStub("AceDB-3.0");


-- #################################################################################################


local function getProfileOption(info)
	return AbsorbsMonitor.db.profile[info.arg];
end

local function setProfileOption(info, value)
	AbsorbsMonitor.db.profile[info.arg] = value;
end


-- #################################################################################################


local defaults = {
	profile = {
		lowValueTolerance = 10,
		priest_globalFactor = 1.3,
	},
};


-- #################################################################################################


local options = {
	type = 'group',
	get = getProfileOption,
	set = setProfileOption,
	
	args = {
		general = {
			type = "group",
			name = "General Options",
			cmdInline = true,
			order = -1,
			
			args = {
				enabled = {
					type = "toggle",
					order = 1,
					name = "Enabled",
					desc = "Enable/Disable this addon",
					width = "double",
					get = function() return AbsorbsMonitor:IsEnabled() end,
					set = function()
						if(AbsorbsMonitor:IsEnabled()) then
							AbsorbsMonitor:Disable();
						else
							AbsorbsMonitor:Enable()
						end
					end,
				},
				options = {
					type = "execute",
					order = 2,
					name = "Options Window",
					desc = "Shows the Options Window",
					func = function()
						AbsorbsMonitor:ShowOptions();
					end,
					guiHidden = true,
				},
				lowcap = {
					type = "range",
					min = 0,
					max = 10000,
					softMin = 0,
					softMax = 250	,
					name = "Low Value Tolerance",
					desc = "Accepted tolerance for underestimating the absorb value",
					arg = "lowValueTolerance",
					
				},
			},
		},
		
		scaling = {
			type = "group",
			name = "Scaling",
			desc = "Options for spell scaling and modifiers",
			guiHidden = true, -- only temporary
			
			args = {
				unit = {
					type = "input",
					name = "Unit Scaling",
					desc = "Prints public scaling information by UnitId",		
					get = function() return 0; end,
					set = function(info, value)
						local guid = UnitGUID(value);
						
						if(not guid) then
							AbsorbsMonitor:Print("Specify a valid UnitId like target, focus, party4, raid3 that can be resolved to a GUID");
						else
							AbsorbsMonitor:PrintScaling(guid);
						end
					end,
					guiHidden = true,
				},
				player = {
					type = "execute",
					name = "Player Scaling",
					desc = "Prints public and private scaling information for you",
					func = function()
						AbsorbsMonitor:PrintPlayerScaling();
					end,
					guiHidden = true,
				},
				
			},
		},
		
		priest = {		
			type = "group",
			name = "Priest",
			desc = "Priest-specific options",
			
			args = {			
				globalmod = {
					type = "range",
					min = 0.0,
					max = 10.0,
					softMin = 0.9,
					softMax = 1.3,
					name = "Global Factor",
					desc = "A constant factor to multiply all Power Word: Shield values with",
					arg = "priest_globalFactor",
				},				
			},			
		},
	},
};


-- #################################################################################################


function AbsorbsMonitor:InitOptions()
	self.db = AceDB:New("AbsorbsMonitorDB", defaults);
	
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db);
	options.args.profile.order = -2;
	
	AceConfig:RegisterOptionsTable("AbsorbsMonitor", options, {"am", "absorbsmonitor"});
	
	self.optionsFrames = {}
	self.optionsFrames.general = AceConfigDialog:AddToBlizOptions("AbsorbsMonitor", nil, nil, "general");
	self.optionsFrames.profile = AceConfigDialog:AddToBlizOptions("AbsorbsMonitor", "Profiles", "AbsorbsMonitor", "profile");
end

function AbsorbsMonitor:ShowOptions()
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.profile);
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.general);
end
